@@ -890,10 +890,15 @@ struct hdf5_iterate_info {
 static void
 add_to_list (struct hdf5_iterate_info *iinfo, const char *name, SEXP obj)
 {
-  PROTECT (iinfo->ret);
-  iinfo->ret = CONS (obj, iinfo->ret);
-  TAG (iinfo->ret) = install ((char *) name);
+  SEXP pair;
+  
+  PROTECT (pair = CONS (obj, CDR (iinfo->ret)));
+  TAG (pair) = install ((char *) name);
+  SETCDR (iinfo->ret, pair);
   UNPROTECT (1);
+
+  if (hdf5_global_verbosity > 2)
+    Rprintf ("Adding `%s' to list\n", name);
 }
 
 
@@ -904,29 +909,14 @@ collect (SEXP call, hid_t id, H5G_iterate_t iterate_func, SEXP env)
 
   iinfo.call = call;
   iinfo.add = add_to_list;
-  iinfo.ret = R_NilValue;
+  PROTECT (iinfo.ret = CONS (R_NilValue, R_NilValue));
   iinfo.env = env;
 
   if (H5Giterate (id, ".", NULL, iterate_func, &iinfo) < 0)
     errorcall (call, "unable to collect HDF group");
 
-  {
-    SEXP nl = R_NilValue, l;
-    SEXP rl = iinfo.ret;
-
-    PROTECT (rl);
-    l = rl;
-    while (l != R_NilValue)
-      {
-	PROTECT (nl);
-	nl = CONS (CAR (l), nl);
-	TAG (nl) = TAG (l);
-	UNPROTECT (1);
-	l = CDR (l);
-      }
-    UNPROTECT (1);
-    return nl;
-  }
+  UNPROTECT (1);
+  return CDR (iinfo.ret);
 }
 
 static void
@@ -1179,6 +1169,8 @@ hdf5_process_object (hid_t id, const char *name, void *client_data)
 	errorcall (iinfo->call, "unable to open group `%s'", name);
 
       PROTECT (l = collect (iinfo->call, gid, hdf5_process_object, iinfo->env));
+      if (hdf5_global_verbosity > 2)
+        Rprintf ("Adding `%s'\n", name);
       iinfo->add (iinfo, name, l);
 
       hdf5_load_attributes (iinfo->call, gid, l, name);
@@ -1441,17 +1433,6 @@ add_to_symbol_table (struct hdf5_iterate_info *iinfo,
   setVar (install ((char *)name), obj, iinfo->env);
 }
 
-static void
-add_to_return_list (struct hdf5_iterate_info *iinfo,
-		    const char *name,
-		    SEXP obj)
-{
-  PROTECT (iinfo->ret);
-  iinfo->ret = CONS (obj, iinfo->ret);
-  TAG (iinfo->ret) = install ((char *) name);
-  UNPROTECT (1);
-}
-
 SEXP do_hdf5load (SEXP args)
 {
   const char *path;
@@ -1492,12 +1473,13 @@ SEXP do_hdf5load (SEXP args)
     errorcall (call, "Unable to register string->ref converter");
 
   iinfo.call = call;
-  iinfo.add = restore_syms ? add_to_symbol_table : add_to_return_list;
+  iinfo.add = restore_syms ? add_to_symbol_table : add_to_list;
   iinfo.env = env;
-  iinfo.ret = R_NilValue;
+  PROTECT (iinfo.ret = CONS (R_NilValue, R_NilValue));
 
   if (H5Giterate (fid, "/", NULL, hdf5_process_object, &iinfo) < 0)
     errorcall (call, "unable to iterate over HDF file: %s", path);
 
-  return iinfo.ret;
+  UNPROTECT (1);
+  return CDR (iinfo.ret);
 }
